/*
 * Copyright 2021 European Commission
 *
 * Licensed under the EUPL, Version 1.2 or – as soon they will be approved by the European Commission - subsequent versions of the EUPL (the "Licence");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 *     https://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and limitations under the Licence.
 */
; // jshint ignore:line
define(function leosPluginUtilsModule(require) {
    "use strict";

    var CKEDITOR = require("promise!ckEditor");
    var TEXT = "text";
    var BOGUS = "br";
    var UNKNOWN = "unknown";
    var ORDER_LIST_ELEMENT = "ol";
    var HTML_POINT = "li";
    var MAX_LEVEL_DEPTH = 7;
    var MAX_LIST_LEVEL = 5;
    var MAX_LEVEL_LIST_DEPTH = 4;

    function _hasTextOrBogusAsNextSibling(element){
        return (element instanceof CKEDITOR.dom.element) && element.hasNext()
            && (_getElementName(element.getNext()) === TEXT || _getElementName(element.getNext()) === BOGUS);
    }

    function _getElementName(element) {
        var elementName = UNKNOWN;
        if (element instanceof CKEDITOR.dom.element) {
            elementName = element.getName();
        } else if (element instanceof CKEDITOR.dom.text
                || (element && element.nodeName && element.nodeName === "#text")) {
            elementName = TEXT;
        } else if(element && element.localName){
            elementName = element.localName;
        }
        return elementName;
    }

    function _setFocus(element, editor){
        if(element){
            var range = editor.createRange();
            range.selectNodeContents(element);
            range.collapse(true);
            range.select();
            range.scrollIntoView();
        }
    }

    /**
     * Calculates the depth level of the selected element inside the list (ol block).
     * It counts how many ol elements are present in the hierarchy.
     */
    function _calculateListLevel(selected) {
        var level = 0;
        var actualEL = selected;
        while (_isListElement(actualEL)) {
            level++;
            actualEL = actualEL.getAscendant(ORDER_LIST_ELEMENT);
        }
        return level;
    }

    /**
     * Returns true if the selected element is child of an li or ol
     */
    function _isListElement(el) {
        return (el && (el.getAscendant(ORDER_LIST_ELEMENT) || el.getAscendant(HTML_POINT)));
    }
    
    function _isAnnexList(element) {
        return element && 'aknAnnexList' === element.getAttribute('data-akn-name');
    }
    
    function _isSelectionInFirstLevelList(element) {
        var listDepth = 0;
        var olElement = element.getAscendant(ORDER_LIST_ELEMENT);
        while (!_isAnnexList(olElement)) {
            if (olElement) {
                listDepth = listDepth + 1;
                olElement = olElement.getAscendant(ORDER_LIST_ELEMENT);
            } else {
                break;
            }
        }
        return listDepth === 1;
    }
    
    function _getAnnexList(element) {
        var olElement = element.getAscendant(ORDER_LIST_ELEMENT);
        while (olElement && !_isAnnexList(olElement)) {
            olElement = olElement.getAscendant(ORDER_LIST_ELEMENT);
        }
        return olElement;
    }

    return {
        hasTextOrBogusAsNextSibling: _hasTextOrBogusAsNextSibling,
        getElementName: _getElementName,
        setFocus: _setFocus,
        calculateListLevel: _calculateListLevel,
        getAnnexList: _getAnnexList,
        isSelectionInFirstLevelList: _isSelectionInFirstLevelList,
        isAnnexList: _isAnnexList,
        MAX_LEVEL_DEPTH: MAX_LEVEL_DEPTH,
        MAX_LIST_LEVEL: MAX_LIST_LEVEL,
        MAX_LEVEL_LIST_DEPTH: MAX_LEVEL_LIST_DEPTH
    };
});
